/* panama_collision.c */

/**************************************************************************+
*
*  Code for generating colliding messages for PANAMA, as specified in 
*  the paper 'Generating Collisions for Panama, instantaneously' by Joan 
*  Daemen and Gilles Van Assche, presented at the Fast Software Encryption 
*  Workshop, Luxembourg, 2007, see "Proceedings of FSE 2007", edited by 
*  Alex Biryukov, Springer-Verlag, to appear.
*  This code is made available on http://radiogatun.noekeon.org/panama
*
*  Code by Joan Daemen
*
*  Notes:  This code is supplied for the purposes of illustrating 
*          the collision generating attack only 
*
+**************************************************************************/


#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#define Obar 0xffffffff

#include "panama.h"

UINT32 ROR(UINT32 a, int offset)
/* intra-word rotations */
{
  UINT32 hlp;
  offset = offset%32;
  if (offset<0) offset += 32;
  return ((a<<offset) ^ (a>>(32-offset)));
}

void printdiffbuffer(UINT32 *B, UINT32 *b)
/* prints 0 (no difference),X (difference) per buffer stage */
{
  int i, j, hlp;
  printf(" ");
   for (i=0 ; i<32 ; i++) {
       hlp=0;
       for (j=0 ; j<8 ; j++){
           if ((B[j+(8*i)]^b[j+(8*i)]) == Obar) hlp = 1;
       } 
       if (hlp==0) printf("-");
       else printf("X"); 
   }
}


void printdiffstate(UINT32 *A, UINT32 *B, UINT32 *a, UINT32 *b)
/* prints 0 (no difference),1 (symmetric difference) or x (otherwise) per state word */
{
   int i, j, hlp;
   for (i=0 ; i<17 ; i++) {
       if ((A[i]^a[i]) == Obar) printf("1"); 
       else if ((A[i]^a[i]) == 0) printf("-");
       else printf("x");
       }
}

void mypanama_copy(UINT32 *A, UINT32 *B, UINT32 *a, UINT32 *b)
/* copies state A and buffer B into state a and buffer b */
{
int i,j;

for (i=0 ; i<17 ; i++) a[i] = A[i];
for (i=0 ; i<32 ; i++) {
    for (j=0 ; j<8 ; j++) b[j+(8*i)] = B[j+(8*i)];}

}

void mypanama_reset(UINT32 *A, UINT32 *B)
/* resets state A and buffer B to all zero */
{
int i,j;

for (i=0 ; i<17 ; i++) A[i] = 0;
for (i=0 ; i<32 ; i++) {
    for (j=0 ; j<8 ; j++) B[j+(8*i)] = 0;
    }

}
 
void mypanama_input(UINT32 *P, UINT32 *A, UINT32 *B)
/* inputs input vector P into state A and buffer B */
{
int i;

for (i=0 ; i<8 ; i++) A[i+1] ^= P[i];
for (i=0 ; i<8 ; i++) B[i]   ^= P[i];

}

void mypanama_round(UINT32 *A, UINT32 *B)
/* applies updating function to state A and buffer B, without input */
{

int i,j;
UINT32 a[17] ; UINT32 b[8*32];

/* the mill */
for (i=0 ; i<17 ; i++) a[i] = ~A[i] ^ (~A[(i+1)%17])&A[(i+2)%17];
for (i=0 ; i<17 ; i++) A[i] = ROR(a[(7*i)%17],(i*(i+1))/2);
for (i=0 ; i<17 ; i++) a[i] = A[i] ^ A[(i+1)%17] ^ A[(i+4)%17];
a[0] ^= 1;

/* belt injection */
for (i=0 ; i<8 ; i++) a[i+9] ^= B[i+(16*8)];

/* the belt */
for (i=0 ; i<31 ; i++) {
    for (j=0 ; j<8 ; j++) b[j+(8*(i+1))] = B[j+(8*i)];
    }
for (j=0 ; j<8 ; j++) b[j+(8* 0)]  = B[j+(8*31)];
for (j=0 ; j<8 ; j++) b[j+(8*25)] ^= B[((j+2)%8)+(8*31)];

/* putting back in A and B*/
for (i=0 ; i<17 ; i++) A[i] = a[i];
for (i=0 ; i<32 ; i++) {
    for (j=0 ; j<8 ; j++) B[j+(8*i)] = b[j+(8*i)];
    }

}

void mypanama_step(UINT32 *P, UINT32 *A, UINT32 *B)
/*  input followed by round, useful in generating the collision */
{
     mypanama_input(P,A,B);
     mypanama_round(A,B);
}

void pre(UINT32 *A, UINT32 *B, UINT32 *a, UINT32 *b, UINT32 *P)
/* a,b is the result of applying step to A,B with input word P */
{
   mypanama_copy(A,B,a,b);
   mypanama_input(P,a,b);
   mypanama_round(a,b);
   return;   
}

void mypanama_generate_difference(UINT32 *D, int offset)
/* generates the difference at the given offset */
{
   int i;
   for (i=0 ; i<8*(offset+36) ; i++) D[i] = 0;
   D = D+8*offset;
   /* 0xa0 */
   D[5+(8*1)] = Obar;
   D[7+(8*1)] = Obar;
   /* 0x0b */
   D[0+(8*2)] = Obar;
   D[1+(8*2)] = Obar;
   D[3+(8*2)] = Obar;
   /* 0xde */
   D[1+(8*3)] = Obar;
   D[2+(8*3)] = Obar;
   D[3+(8*3)] = Obar;
   D[4+(8*3)] = Obar;
   D[6+(8*3)] = Obar;
   D[7+(8*3)] = Obar;

   /* 0x28 */
   D[3+(8*8)] = Obar;
   D[5+(8*8)] = Obar;
   /* 0xc1 */
   D[6+(8*9)] = Obar;
   D[7+(8*9)] = Obar;
   D[1+(8*9)] = Obar;
   /* 0xb7 */
   D[7+(8*10)] = Obar;
   D[0+(8*10)] = Obar;
   D[1+(8*10)] = Obar;
   D[2+(8*10)] = Obar;
   D[4+(8*10)] = Obar;
   D[5+(8*10)] = Obar;

   /* 0xa0 */
   D[5+(8*33)] = Obar;
   D[7+(8*33)] = Obar;
   /* 0x0b */
   D[0+(8*34)] = Obar;
   D[1+(8*34)] = Obar;
   D[3+(8*34)] = Obar;
   /* 0xde */
   D[1+(8*35)] = Obar;
   D[2+(8*35)] = Obar;
   D[3+(8*35)] = Obar;
   D[4+(8*35)] = Obar;
   D[6+(8*35)] = Obar;
   D[7+(8*35)] = Obar;

}

void mypanama_collision_printdiff(UINT32 *D, UINT32 *M, int offset)
/* prints the difference evolution in state and buffer during the collision 
   D: difference between the colliding messages
   M: one of the messages */
{
   int i, j;

   UINT32 AAl[17], ABl[8*32];
   UINT32 AAr[17], ABr[8*32];
   UINT32 P[8];
   
   mypanama_reset(AAl,ABl);
   mypanama_round(AAl,ABl);
   mypanama_reset(AAr,ABr);
   mypanama_round(AAr,ABr);
   for (i=0 ; i<offset+36 ; i++){
       mypanama_input(M+8*i,AAl,ABl);
       for (j=0 ; j<8 ; j++) P[j] = M[j+8*i] ^ D[j+8*i];
       mypanama_input(P,AAr,ABr);

       printf("round %2d  ",i);
       printdiffstate(AAl,ABl,AAr,ABr);
       printdiffbuffer(ABl,ABr);
       printf("\n");       

       mypanama_round(AAl,ABl);
       mypanama_round(AAr,ABr);
       }

}

void mypanama_collision(UINT32 *M, int offset)
/* adapts the message M so that it gives rise to a collision */
{

   UINT32 A[17], B[8*32], a[17], b[8*32], *P, *Q;
   int RC[17] = {-0,-15,-55,-120,-6,-36,-91,-1,-21,-66,-136,-10,-45,-105,-3,-28,-78}; 

   mypanama_reset(A,B);
   mypanama_round(A,B);
   for (P = M ; P < M+8*offset ; P += 8) mypanama_step(P,A,B);
   /* subcollision I   */
   /* round 0          */

   /* T9               */ pre(A,B,a,b,P); P[1] ^= ROR(a[9],RC[2]) ^ Obar; 

   /* round 1          */ mypanama_step(P,A,B); P += 8;

   /* T5               */ P[4] = A[5];
   /* T7               */ P[6] = A[7];
   /* W13              */ pre(A,B,a,b,P); P[5] ^= ROR(a[13],RC[6]) ^ Obar; 
   /* W16 + W0         */ pre(A,B,a,b,P); P[3] ^= ROR(a[16]^a[0],RC[4]) ^ Obar; 
   /* W15              */ pre(A,B,a,b,P); P[2] ^= ROR(a[15],RC[3]) ^ Obar; 

   /* round 2          */ mypanama_step(P,A,B); P += 8;

   /* W1               */ P[0] = A[1];
   /* W2 + W3          */ P[2] = P[1] ^ A[3] ^ A[2] ^ Obar;
   /* W3 + W4          */ P[3] = P[2] ^ A[4] ^ A[3];
   /* W4 + W5          */ P[4] = P[3] ^ A[5] ^ A[4];
   /* W5 + W6          */ P[5] = P[4] ^ A[6] ^ A[5] ^ Obar;
   /* W7               */ P[6] = A[7] ^ Obar;

   for (; P < M+8*(offset+7) ; P += 8) mypanama_step(P,A,B);
   /* subcollision II  */
   /* round 7          */
   
   /* Q1  dep. removal */ P[0] = A[1] ^ Obar;  
   /* T11 dep. removal */ pre(A,B,a,b,P); P[2] ^= ROR(a[11],RC[3]);

                          pre(A,B,a,b,P); Q = P+8;
                          Q[2] = a[3];
                          Q[4] = a[5];
                          Q[6] = a[7];
                          Q[7] = a[8] ^ Obar; 
                          Q[0] = a[1]; 
                          mypanama_step(Q,a,b);
   /* W14 via T13      */ P[5] ^= ROR(a[14],RC[6]+RC[13]);
   
                          pre(A,B,a,b,P); Q = P+8;
                          Q[2] = a[3];
                          Q[4] = a[5];
                          Q[6] = a[7];
                          Q[7] = a[8] ^ Obar; 
                          Q[0] = a[1]; 
                          mypanama_step(Q,a,b);
   /* W11 via T9       */ P[1] ^= ROR(a[11],RC[2]+RC[9]); 
   
   /* round 8          */ mypanama_step(P,A,B); P += 8;

   /* T3               */ P[2] = A[3];  
   /* T5               */ P[4] = A[5];  
   /* T7               */ P[6] = A[7];  
   /* T8  dep. removal */ P[7] = A[8] ^ Obar; 
   /* W12 + W13 side   */ P[0] = A[1];                           
   /* W12 + W13 side   */ pre(A,B,a,b,P); P[1] ^= ROR(a[12]^a[13],RC[0]); 
   /* W9               */ pre(A,B,a,b,P); P[5] ^= ROR(a[9],RC[6]); 

   /* round 9          */ mypanama_step(P,A,B); P += 8;

   /* W1               */ P[0] = A[1];
   /* W3               */ P[2] = A[3];
   /* W5               */ P[4] = A[5];
   /* W6 + W7          */ P[6] = P[5] ^ A[6] ^ A[7];
   /* W7 + W8          */ P[7] = P[6] ^ A[7] ^ A[8] ^ Obar;

   for (; P < M+8*(offset+17) ; P += 8) mypanama_step(P,A,B);
   /* subcollision III */
   /* round 17         */

   /* T0               */ pre(A,B,a,b,P); P[6] ^= ROR(a[0],RC[7]) ^ Obar; 
   /* T13              */ pre(A,B,a,b,P); P[5] ^= ROR(a[13],RC[6]); 
   /* T15              */ pre(A,B,a,b,P); P[2] ^= ROR(a[15],RC[3]); 

   /* round 18         */ mypanama_step(P,A,B); P += 8;

   /* W12              */ pre(A,B,a,b,P); P[5] ^= ROR(a[12],RC[6]) ^ Obar; 
   /* T4 dep. removal  */ P[3] = A[4];  
   /* W9               */ pre(A,B,a,b,P); P[1] ^= ROR(a[9],RC[2]); 
   /* W10 + W11        */ pre(A,B,a,b,P); P[2] ^= ROR(a[10]^a[11],RC[3]); 

   /* round 19         */ mypanama_step(P,A,B); P += 8;

   /* W2               */ P[1] = A[2];
   /* W4               */ P[3] = A[4] ^ Obar;
   /* W5               */ P[5] = P[4] ^ A[5] ^ A[6] ^ Obar;
   /* W6               */ P[6] = P[5] ^ A[6] ^ A[7] ^ Obar;
   /* W8               */ P[7] = A[8];

   for (; P < M+8*(offset+23) ; P += 8) mypanama_step(P,A,B);
   /* subcollision IV  */
   /* round 23         */

   /* T11 side via Q0  */ pre(A,B,a,b,P); P[6] ^= ROR(a[0],RC[7]);

   /* round 24         */ mypanama_step(P,A,B); P += 8;

   /* T2  dep. removal */ P[1] = A[2]; 
   /* T13              */ pre(A,B,a,b,P); P[5] ^= ROR(a[13],RC[6]) ^ Obar; 
   /* T15              */ pre(A,B,a,b,P); P[2] ^= ROR(a[15],RC[3]); 
   /* T11 side         */ pre(A,B,a,b,P); P[0] ^= ROR(a[11],RC[16]);  
   
   /* round 25         */ mypanama_step(P,A,B); P += 8;

   /* T8  dep. removal */ P[7] = A[8];
   /* W9               */ pre(A,B,a,b,P); P[5] ^= ROR(a[9]^a[10],RC[6]) ^ Obar;
   /* W13 side         */ P[1] = A[2] ^ Obar;
   /* W13 side         */ pre(A,B,a,b,P); P[0] ^= ROR(a[13],RC[0]);
   /* W16              */ pre(A,B,a,b,P); P[3] ^= ROR(a[16],RC[4]);
   /* W11              */ pre(A,B,a,b,P); P[2] ^= ROR(a[11],RC[3]);
   /* W14              */ pre(A,B,a,b,P); P[6] ^= ROR(a[14],RC[7]) ^ Obar;

   /* round 26         */ mypanama_step(P,A,B); P += 8;

   /* W0 + W1          */ P[0] = A[0] ^ A[1];
   /* W1 + W2          */ P[1] = P[0] ^ A[1] ^ A[2];
   /* W2 + W3          */ P[2] = P[1] ^ A[2] ^ A[3];
   /* W3 + W4          */ P[3] = P[2] ^ A[3] ^ A[4] ^ Obar;
   /* W4 + W5          */ P[4] = P[3] ^ A[4] ^ A[5] ^ Obar;
   /* W6               */ P[5] = A[6] ^ Obar;
   /* W7               */ P[6] = A[7] ^ Obar;
   /* W8 + W9          */ P[7] = A[8] ^ A[9] ^ Obar;
   
   for (; P < M+8*(offset+32) ; P += 8) mypanama_step(P,A,B);
   /* subcollision V   */
   /* round 32         */
   
   /* T9               */ pre(A,B,a,b,P); P[1] ^= ROR(a[9],RC[2]) ^ Obar; 

   /* round 33         */ mypanama_step(P,A,B); P += 8;

   /* T5               */ P[4] = A[5];
   /* T7               */ P[6] = A[7];
   /* W13              */ pre(A,B,a,b,P); P[5] ^= ROR(a[13],RC[6]) ^ Obar; 
   /* W16 + W0         */ pre(A,B,a,b,P); P[3] ^= ROR(a[16]^a[0],RC[4]) ^ Obar; 
   /* W15              */ pre(A,B,a,b,P); P[2] ^= ROR(a[15],RC[3]) ^ Obar; 

   /* round 34         */ mypanama_step(P,A,B); P += 8;

   /* W1               */ P[0] = A[1];
   /* W2 + W3          */ P[2] = P[1] ^ A[3] ^ A[2] ^ Obar;
   /* W3 + W4          */ P[3] = P[2] ^ A[4] ^ A[3];
   /* W4 + W5          */ P[4] = P[3] ^ A[5] ^ A[4];
   /* W5 + W6          */ P[5] = P[4] ^ A[6] ^ A[5] ^ Obar;
   /* W7               */ P[6] = A[7] ^ Obar;

}
